/* Copyright 2005 Scott Torborg */

/* This code is released under the GNU General Public License. 
 * For more information, see the COPYING file included in this 
 * package. */

/* DISCO DANCE FLOOR MAIN CONTROL LIBRARIES
 * First East, MIT
 * March 2005 */

/* TO DO:
 * module/floor reading from touch sensors
 * device autodetection
 */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <termios.h>
#include <assert.h>    
#include <stdlib.h>
#include <limits.h>
#include <pthread.h>

#include "ddf.h"

/* pretty printing of errors */
int
check(int err, char *s)
{
	if (err < 0)
	    fprintf(stderr,"ERROR! %s: %s\n", s, strerror(errno));
#ifdef DEBUG
	else
	    fprintf(stderr,"OK %s\n", s);
#endif
	
	fflush(stderr);
	return err;
}

/** Takes a bitmap in RGB888 format and changes it to an RRR...GGG...BBB...  nib
 * format. The low-order nib in each byte is written to first. The data is taken
 * from the high-order nib of each byte */
unsigned char*
mash_data(unsigned char* in_buffer, int buffer_len) { 
	int tmp, n_index, i;
	char* mashed;
	/* we have to be passed a multiple of 3 and 2 */	

	if (buffer_len%3 != 0 || buffer_len%2 != 0) {
		fprintf(stderr, "ERROR! buffer_len must be a multiple of both two and three!\n");
		return NULL;
	}
	
	/* don't forget to free me when we're done */
	mashed = malloc(buffer_len / 2 + 1);
	memset(mashed, 0x00, buffer_len / 2 + 1);
	
	mashed[0] = 0x10;
	
	int row;
	
	// R0 1b,  G0 1b,  B0 1b ... = 1 row (48 bytes)
	// R1 1n,  R0 1n,  R3 1n, (to 16, then g, b) =  1 row (24 bytes)
	
	for (i = 0; i < buffer_len; i++) {
		int row = i / 48;
		int color = i % 3;
		int index = (i / 3) % 16;
		
		int byte_offset = (row * 3 + color) * 8;
		
		unsigned char val = 0xFF - in_buffer[i];
		
		if (index % 2) {
			mashed[byte_offset + index/2] |= val & 0xF0;
		} else {
			mashed[byte_offset + index/2] |= (val >> 4) & 0xF;
		}
	}
	return mashed;
}

/* write_dancefloor with similar semantics to the old code */
int
write_dancefloor(dancefloor* fl, unsigned char* dancefloor_data)
{
	memcpy(fl->buf,dancefloor_data,ROWS*COLS*3);
	write_dancefloor_buf(fl);
	return 0;
}

/* write_dancefloor writes LED data to the entire dancefloor, wrapping write_module */            
int
write_dancefloor_buf(dancefloor* fl)
{
	wait_for_op(fl);
	send_signal(fl,WRITE_SIGNAL);
	return 0;
}

/* reinit the whole dancefloor */
int
reinit_dancefloor(dancefloor* fl)
{
	wait_for_op(fl);
	send_signal(fl,REINIT_SIGNAL);
	return 0;
}

/* power down the whole dancefloor */
int
powerdown_dancefloor(dancefloor* fl)
{
	wait_for_op(fl);
	send_signal(fl,POWERDOWN_SIGNAL);
	return 0;
}

/* write_module writes LED data to an entire 64-tile module */
int
write_module(int fd, unsigned char *mod_data)
{
	if(fd < 0) {
		return ERROR_W;
	}
	
	unsigned char cmd[OUTPUT_MODULE_BYTES+1];
	unsigned char status[1];
	int i;
	
	/* flush the serial port buffer */
	tcflush(fd, TCIOFLUSH);
	
	/* convert the data to hardware format and append it to cmd */
	cmd[0] = MODULE_WRITE_CODE;
	
	memcpy(cmd+1, mash_data(mod_data, OUTPUT_MODULE_BYTES*2), OUTPUT_MODULE_BYTES);
	
	/* write the data out the serial port and confirm transmission */
	i = write(fd, cmd, OUTPUT_MODULE_BYTES+1);
	if (i != OUTPUT_MODULE_BYTES+1) {
		fprintf(stderr,"ERROR write_module (%d): writing data out (%d)\n",fd,i);
		return ERROR_W;
	}
	
	/* read the status response byte */
	i = read(fd, status, 1);
	if (i != 1) {
		/* serial port read failure */
		fprintf(stderr,"ERROR write_module (%d):  reading result (%d)\n", fd, i);
		return ERROR_R;
	}
	/* check to be sure the status byte is correct */
	if(status[0] != SUCCESS_CODE) {
		fprintf(stderr,"ERROR write_module (%d): write unsuccessful (%x)\n", fd, status[0]);
		return ERROR_W;
	}
	
	/* if we make it this far, return success */
	return SUCCESS_CODE;
}

/* reinit turns off every LED, although the module stays
 * awake */
int
reinit_module(int fd)
{
	if(fd < 0) {
		return ERROR_W;
	}
	
	unsigned char b[1];
	b[0] = REINIT_CODE;
	
	/* write the power down byte */
	check(write(fd, b, 1), "write cmd: reinitialize");
	
	/* read the status response */
	check(read(fd, b, 1), "read result");
	if(b[0] != SUCCESS_CODE) {
		fprintf(stderr,"ERROR! reinitialize unsuccessful: %x\n", b[0]);
		return -1;
	}
	return 1;
}

/* power_down turns off every LED, although the module stays
 * awake */
int
powerdown_module(int fd)
{
	if(fd < 0) {
		return ERROR_W;
	}
	
	unsigned char b[1];
	b[0] = POWER_DOWN_CODE;
	
	/* write the power down byte */
	check(write(fd, b, 1), "write cmd: power down");
	
	/* read the status response */
	check(read(fd, b, 1), "read result");
	if(b[0] != SUCCESS_CODE) {
		fprintf(stderr,"ERROR! power down unsuccessful: %x\n", b[0]);
		return -1;
	}
	return 1;
}

/* look for device names, setup serial ports, probe modules
 * for echo, and add modules to the pool if they are
 * fully operational. */
dancefloor*
init(void)
{
	/* allocate dancefloor structure */
	dancefloor* fl = (dancefloor *) malloc(sizeof(dancefloor));
	/* allocate dancefloor data buffer */
	fl->buf = malloc(ROWS*COLS*3);
	/* allocate device name table */
	char **device_names = malloc(MODULES_PER_FLOOR*_POSIX_PATH_MAX);
	
	/* temporary thread container */
	pthread_t temp_thread;
	
	/* allocate operation counter */
	fl->modules_done = malloc(sizeof(int));
	/* operation counter mutex, gets passed to dancefloor and all modules */
	fl->count_mut = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(fl->count_mut, NULL);
	/* modules done condition variable, gets passed to both */
	fl->done = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
	pthread_cond_init(fl->done, NULL);
	
	/* allocate signal specifier */
	fl->signal_spec = malloc(sizeof(int));
	/* signal mutex gets passed to both */
	fl->signal_mut = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(fl->signal_mut, NULL);
	/* signal condition variable, gets passed to both */
	fl->signal_cond = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
	pthread_cond_init(fl->signal_cond, NULL);
	
	/* set up platform-specific device name table */
	device_names[0] = "/dev/tty.usbserial-1B111";
	device_names[1] = "/dev/tty.usbserial-1B112";
	device_names[2] = "/dev/tty.usbserial-1B113";
	device_names[3] = "/dev/tty.usbserial-1B114";
	device_names[4] = "/dev/tty.usbserial-1B121";
	device_names[5] = "/dev/tty.usbserial-1B122";
	device_names[6] = "/dev/tty.usbserial-1B123";
	device_names[7] = "/dev/tty.usbserial-1B124";
	
	/* iterator */
	int modnum;
	
	/* get a lock on the operations counter */
	pthread_mutex_lock(fl->count_mut);
	/* reset the counter */
	*fl->modules_done = 0;
	/* unlock the operations counter */
	pthread_mutex_unlock(fl->count_mut);
	
	for(modnum = 0; modnum < MODULES_PER_FLOOR; modnum++) {
		/***** BEGIN ****/
		
		/* allocate module struct */
		moduleinfo *mod;
		mod = malloc(sizeof(moduleinfo));
		/* assign instance-specific vars */
		mod->name = device_names[modnum];
		mod->num = modnum;
		
		/* pass buffer pointer */
		mod->buf = fl->buf;
		
		/* pass operations counter mutex */
		mod->count_mut = fl->count_mut;
		/* pass operations counter */
		mod->modules_done = fl->modules_done;
		/* pass modules_done condition variable */
		mod->done = fl->done;
		
		/* pass signal specifier */
		mod->signal_spec = fl->signal_spec;
		/* pass signal mutex */
		mod->signal_mut = fl->signal_mut;
		/* pass signal condition variable */
		mod->signal_cond = fl->signal_cond;
				
		/* need to pass each thread struct with
		 * device name, write ready condition
		 * variable, and  buffer area */
		check(pthread_create(&temp_thread, NULL, threadfunc, mod),"creating thread\n");
		/***** END ****/
	}
	
	/* don't need the device nametable anymore */
	free(device_names);
			
	/* return dancefloor struct */
	return fl;
}

/* each thread executes this function */
void*
threadfunc(void *modp)
{
	/* cast moduleinfo from void pointer */
	moduleinfo* mod = (moduleinfo*) modp;
	char *name;
	int fd, bitrate, x, err;
	struct termios t;
	
	/* select baud rate */
	switch (BAUD_RATE) {
		case 4800: bitrate = B4800; break;
		case 9600: bitrate = B9600; break;
		case 19200: bitrate = B19200; break;
		case 38400: bitrate = B38400; break;
		case 57600: bitrate = B57600; break;
		case 115200: bitrate = B115200; break;
		default:
			fprintf(stderr, "Unknown bitrate %d\n", bitrate);
			exit(1);
	}
		
	for(x = 1; x < 2; x++) {
		printf("opening port %s\n", mod->name);
		
		/* open the serial device */
		fd = open (mod->name, O_RDWR, 0);
		if (fd < 0) {
			fprintf (stderr, "%s: %s\n", mod->name, strerror (errno));
			close (fd);
			fd = -1;
			break;
		}
		
		/* get the serial port address */
		err = tcgetattr (fd, &t);
		if (err < 0) {
			fprintf (stderr, "tcgetattr: %s\n", strerror (errno));
			close (fd);
			fd = -1;
			break;
		}
		
		/* set the baud rate */
		err = cfsetispeed (&t, bitrate);
		if (err < 0) {
			fprintf (stderr, "cfsetispeed: %s\n", strerror (errno));
			close (fd);
			fd = -1;
			break;
		}
		err = cfsetspeed (&t, bitrate);
		if (err < 0) {
			fprintf (stderr, "cfsetspeed: %s\n", strerror (errno));
			close (fd);
			break;
			fd = -1;
		}
		
#ifdef __MACOSX__
		/* to use hardware flow control with RTS and CTS
		 * as of 1/28/05 the firmware does not support this */
		if (FLOW_CONTROL) {
			t.c_cflag |= (CRTS_IFLOW | CCTS_OFLOW);
		} else {
			t.c_cflag |= CLOCAL;
			t.c_cflag &= ~(CRTS_IFLOW | CCTS_OFLOW);
		}
		t.c_lflag &= ~ICANON; // USB key driver turns this on by default    
		err = tcsetattr (fd, TCSAFLUSH, &t);
		if (err < 0) {
			fprintf (stderr, "tcsetattr: %s\n", strerror (errno));
			close(fd);
			break;
			fd = -1;
		}
#endif
		
		/* flush the buffer */
		tcflush(fd, TCIOFLUSH);
		
		/* make command byte */
		unsigned char b[1];
		b[0] = ECHO_CODE;
		
		/* write the echo request byte */
		check(write(fd, b, 1), "write cmd: echo test");
		
		/* read the status response */
		check(read(fd, b, 1), "read result");
		if(b[0] != SUCCESS_CODE) {
			fprintf(stderr,"ERROR! echo request failed: %x\n", b[0]);
			close(fd);
			fd = -1;
		}
	}
	
	/* get a lock on the signal mutex */
	pthread_mutex_lock(mod->signal_mut);
	
	/* increment the operations counter */
	pthread_mutex_lock(mod->count_mut);
	*mod->modules_done = *mod->modules_done + 1;
	pthread_mutex_unlock(mod->count_mut);
	pthread_cond_signal(mod->done);
	
	/* local signal */
	int mysig;
	
	/* enter normal operation */
	while(1) {
		do {
			/* block until we get a signal */
			pthread_cond_wait(mod->signal_cond, mod->signal_mut);
			/* loop if condition is zero */
		} while(*mod->signal_spec == 0);
		/* copy signal */
		mysig = *mod->signal_spec;
		/* unlock the signal mutex */
		pthread_mutex_unlock(mod->signal_mut);
				
		/* do shit */
		switch(mysig) {
			/* exit self (should cleanup fd and whatnot first */
			case EXIT_SIGNAL:
				/* close my device */
				close(fd);
								
				/* exit with no particular exit status */
				pthread_exit(NULL);
				break;
			
			/* should call write_module and whatnot */
			case WRITE_SIGNAL:
				/* write my module */
				write_module(fd, mod->buf+(mod->num*OUTPUT_MODULE_BYTES*2));				
				break;
				
			/* should call read_module and whatnot */
			case READ_SIGNAL:
				break;
			
			/* power down my module */
			case POWERDOWN_SIGNAL:
				powerdown_module(fd);
				break;
			
			/* reinit my module */
			case REINIT_SIGNAL:
				reinit_module(fd);
				break;
		}

		/* get a lock on the signal mutex */
		pthread_mutex_lock(mod->signal_mut);
		
		/* increment the counter */
		pthread_mutex_lock(mod->count_mut);
		*mod->modules_done = *mod->modules_done + 1;
		pthread_mutex_unlock(mod->count_mut);
		pthread_cond_signal(mod->done);
		
		/* do the loop da loop! */
	}
	
	return;
}

/* close all the file descriptors */
int
cleanup(dancefloor* fl)
{
	/* wait for any operations to finish */
	wait_for_op(fl);
	/* tell all threads to close devices
	 * and exit themselves */
	send_signal(fl, EXIT_SIGNAL);
	
	/* free all the shit we malloc'd back in init() */
	free(fl->signal_cond);
	free(fl->signal_mut);
	free(fl->signal_spec);
	free(fl->done);
	free(fl->count_mut);
	free(fl->modules_done);
	free(fl->buf);
	free(fl);
	
	return 0;
}

/* wait for any previous operation to finish */
void
wait_for_op(dancefloor *fl)
{
	/* lock op counter */
	pthread_mutex_lock(fl->count_mut);
	while(*fl->modules_done < MODULES_PER_FLOOR) {
		/* modules not all done */
		pthread_cond_wait(fl->done,fl->count_mut);
	}
	/* reset the counter */
	*fl->modules_done = 0;
	/* unlock it */
	pthread_mutex_unlock(fl->count_mut);

	return;
}

/* send an arbitrary int signal */
void
send_signal(dancefloor *fl, int sig)
{
	pthread_mutex_lock(fl->signal_mut);
	*fl->signal_spec = sig;
	pthread_mutex_unlock(fl->signal_mut);
	pthread_cond_broadcast(fl->signal_cond);
}
