#ifndef __DDF_H_
#define __DDF_H_

/* Copyright 2005 Scott Torborg */

/* This code is released under the GNU General Public License. 
 * For more information, see the COPYING file included in this 
 * package. */

/* DISCO DANCE FLOOR CONTROL
 * First East, MIT
 * January 2005 */

#include <pthread.h>

/* set this symbol for debugging */
//#define DEBUG

/* change this to compile on linux, obviously */
#define __MACOSX__

#define BAUD_RATE 57600

/* only applicable on OS X */
#define FLOW_CONTROL 0

/* firmware control codes */
#define SUCCESS_CODE				0x00
#define MODULE_WRITE_CODE		0x10
#define MODULE_READ_CODE		0x20
#define POWER_DOWN_CODE			0x40
#define ECHO_CODE						0x50
#define REINIT_CODE					0x60

/* error codes that write commands can return */
#define ERROR_R				(-1)
#define ERROR_RW 			(-2)
#define ERROR_W				(-3)
#define ERROR_THREAD	(-4)

#define WRITE_SIGNAL			1
#define EXIT_SIGNAL				2
#define READ_SIGNAL				3
#define POWERDOWN_SIGNAL	4
#define REINIT_SIGNAL			5

/* output bytes are the size of the data to be output AFTER
 * being converted to nib-sized hardware format */

#define TILES_PER_ROW      16
#define ROWS_PER_MODULE    4
#define MODULES_PER_FLOOR  8

#define COLS TILES_PER_ROW
#define ROWS ROWS_PER_MODULE*MODULES_PER_FLOOR

static const int OUTPUT_ROW_BYTES = (TILES_PER_ROW * 3) / 2;
static const int SENSOR_ROW_BYTES = TILES_PER_ROW / 8;

static const int OUTPUT_MODULE_BYTES = ((TILES_PER_ROW*3)/2) * ROWS_PER_MODULE;
static const int SENSOR_MODULE_BYTES = (TILES_PER_ROW/8) * ROWS_PER_MODULE;

static const int OUTPUT_FLOOR_BYTES = ((TILES_PER_ROW*3)/2)*ROWS_PER_MODULE*MODULES_PER_FLOOR;
static const int SENSOR_FLOOR_BYTES = (TILES_PER_ROW/8)*ROWS_PER_MODULE*MODULES_PER_FLOOR;

/* an instance of this gets passed to each thread */
typedef struct {
	/* module name, specific to instance */
	char* name;
	/* module number, specific to instance */
	int num;
	/* shared by all */
	unsigned char* buf;
	
	/* modules done condition variable, gets passed to both */
	pthread_cond_t* done;
	/* operation counter mutex, gets passed to both */
	pthread_mutex_t* count_mut;
	/* operation counter, gets passed to both */
	int* modules_done;
	
	/* signal specifier gets passed to both */
	int* signal_spec;
	/* signal mutex gets passed to both */
	pthread_mutex_t* signal_mut;
	/* signal condition gets passed to both */
	pthread_cond_t* signal_cond;
		
} moduleinfo;

/* an instance of this represents the dancefloor
 * and gets passed to the write_dancefloor functions */
typedef struct {
	/* shared by all */
	unsigned char* buf;
	
	/* modules done condition variable, gets passed to both */
	pthread_cond_t* done;
	/* operation counter mutex, gets passed to dancefloor and all modules */
	pthread_mutex_t* count_mut;
	/* operation counter, gets passed to both */
	int* modules_done;
	
	/* signal specifier gets passed to both */
	int* signal_spec;
	/* signal mutex gets passed to both */
	pthread_mutex_t* signal_mut;
	/* signal condition gets passed to both */
	pthread_cond_t* signal_cond;
		
} dancefloor;

/* pretty printing of errors */
int check(int err, char *s);

/* Takes a bitmap in RGB888 format and changes it to an RRR...GGG...BBB...  nib
 * format. The low-order nib in each byte is written to first. The data is taken
 * from the high-order nib of each byte of the input buffer */
unsigned char* mash_data(unsigned char* in_buffer, int buffer_length);

/* write_dancefloor with similar semantics to the old code */
int write_dancefloor(dancefloor* fl, unsigned char* dancefloor_data);

/* write_dancefloor_buf writes fl->buf to the whole dancefloor */
int write_dancefloor_buf(dancefloor* fl);

/* power down the whole dancefloor */
int powerdown_dancefloor(dancefloor* fl);

/* reinit the whole dancefloor */
int reinit_dancefloor(dancefloor* fl);

/* write_module writes LED data to an entire 64-tile module */
int write_module(int fd, unsigned char* mod_data);

/* power_down turns off every LED, although the module stays awake */
int powerdown_module(int fd);

/* reinitialize a module */
int reinit_module(int fd);

/* setup serial ports and return an array of pthreads */
dancefloor* init(void);

/* each thread runs this function */
void* threadfunc(void *device);

/* close all the pthreads */
int cleanup(dancefloor* fl);

/* wait for any previous operation to finish */
void wait_for_op(dancefloor* fl);

/* send arbitrary signal */
void send_signal(dancefloor* fl, int sig);

#endif /* __DDF_H_ */
